	PAGE
;********************************************************
;*		EXPRESSION PARSER			*
;*							*
;*	Gets the value of an expression (a 16 bit #)	*
;* by scanning a buffer.  The buffer is pointed to on	*
;* entry by HL, with max length in C.  Value is 	*
;* returned pushed on the VSTACK.			*
;*							*
;********************************************************
;
;	The syntax of an expression is show below.  The
;grammar is chosen to reflect the parser construction.
;The parser is a top-down machine, so the grammer
;is a Q-grammer with 16 bit integers as attibutes (see
;Compiler Design Theory, by Lewis, Rosencrantz & Stearns)
;
;<expr> ::= <andexpr> <or.list>
;<or.list> ::= OR <and.expr> <or.list>
;		| XOR <and.expr> <or.list>
;		| <null> 
;<and.expr> ::= <not.expr> <and.list>
;<and.list> ::= AND <not.expr> <and.list> 
;		| <null>
;<not.expr> ::= NOT <non-letter> <rel.expr>
;		| <rel.expr>
;<rel.expr> ::= <add.expr> <rel.part>
;<rel.part> ::=   >= <add.expr>
;		| > <add.expr>
;		| <= <add.expr>
;		| <  <add.expr>
;		| != <add.expr>
;		| =  <add.expr>
;		| <null>
;<add.expr> ::= <mult.expr> <add.list>
;<add.list> ::= + <mult.expr> <add.list>
;		| - <mult.expr> <add.list>
;		| <null>
;<mult.expr> ::= <sgd.prim> <mult.list>
;<mult.list> ::= * <sgd.prim> <mult.list>
;		| / <sgd.prim> <mult.list>
;		| MOD <sgd.prim> <mult.list>
;		| SHL <sgd.prim> <mult.list>
;		| SHR <sgd.prim> <mult.list>
;		| <null>
;<sgd.prim> ::= + <primary>
;		| - <primary>
;		| <primary>
;<primary> ::= ( <expr> )
;		| <short.str>
;		| $
;		| LEN[ <expr.list> ]
;		| NULL[  <parm.list> ]
;		| COUNT[  <parm.list> ]
;		| <symbol>
;		| <number>
;<number> ::=  <digits> <base>
;<digits> ::= <digit> <digits> | <null>
;<digit> ::= 0 | 1 | 2 | 3 ...9 | A | B ... F
;<base> ::= B | D | H | Q | <null>
;
;	There are a few interesting problems with the
;grammar given above.  First, some of the operators can be
;prefixes of legal subexpressions - specifically, the 'NOT'
;operator is a prefix of the <add.expr> nonterminal, which can
;generate a <symbol>, which is just a string of letters.
;Thus NOTICE is parsed as NOT ICE.  The <not.expr> parser
;must do one character lookahead to see if the 'NOT' it sees
;is delimited by a non-letter.  In the grammar I have put in
;a special <non-letter> non-terminal which is magical in that
;it is not eaten when recognized!
;
;	The other prefix problems are insignificant.  'AND',
;'OR' and so on can be ripped out of a symbol name if that
;symbol is the second in a sequence of two:  A ANDB comes
;out as A AND B.  This is ok since A ANDB is not syntactical,
;so we have a form of error correction!
;
;	A few backup problems have been encountered:
;The <xx.numb> rules in <primary> are numeric constants that
;are not easily parsable with the Q grammer because they 
;have a base indicator character (H for hex...) FOLLOWING
;the digits of the number:  thus backup is required.  The
;parser rule for numbers handles them in an ad-hoc manner.
;Also in <primary>, the <short.str> thing is a call to the
;<str.expr> parser that checks the length of the returned
;list of integers.  The list must be less than 3 in length.
;
EXPR?	CALL	EXPR
	JMP	???
;
EXPR	XRA	A		;Clear flag, nest level.
	STA	NESTLVL
	STA	BADTERM
	CALL	STKMARK		;Fancy.  If errors, we can
	PUSH	D		;(Save old mark)
	CALL	EXPR1		;undo stack to same spot,
	JC	EXPR11		;and put a zero there.
	POP	D		;Get old mark.
	JMP	STKUNMARK	;Do a recursive unmark.
EXPR11	XCHG
	CALL	STKBACK		;This guarantees that we
	POP	D		;(Unrecurse mark)
	CALL	STKUNMARK	;always return one element
	CALL	PRMZERO 	;on the stack.
	XCHG			;Restore error code in DE.
	STC
	RET
;
;
EXPR1	CALL	ANDEXPR		;Start of actual parsing.
	RC
;
ORLIST	CALL	BREAK
	CALL	TEST
	DB	'OR',0
	JC	ORL2
	CALL	ANDEXPR
	RC
	CALL	STKOR		;'OR' top two elements.
	RC
	JMP	ORLIST
ORL2	CALL	TEST
	DB	'XOR',0
	JC	SUCCEED		;OR and XOR ops optional.
	CALL	ANDEXPR
	RC
	CALL	STKXOR
	RC
	JMP	ORLIST
;
;
ANDEXPR CALL	NOTEXPR
	RC
;
ANDLIST CALL	BREAK
	CALL	TEST
	DB	'AND',0
	JC	SUCCEED
	CALL	NOTEXPR
	RC
	CALL	STKAND		;AND top two stack elements.
	RC
	JMP	ANDLIST
;
;	Weird lookahead of one character is required because
;  of prefix problems: NOTICE = NOT ICE.
;
NOTEXPR CALL	BREAK
	PUSH	H		;Save in case backup required.
	PUSH	B
	CALL	TEST
	DB	'NOT',0
	JC	NOTEX1		;jmp/ don't see a 'NOT'.
	CALL	VARCHAR		;Look ahead one char.
	JNC	NOTEX1		;Jmp/ found another VARCHAR
	POP	D		;Recoginized NOT now.
	POP	D		; so throw away stack.
	CALL	RELEXPR		;Parse a RELEXPR
	RC			;and negate a <rel.expr>.
	JMP	STKNOT		;COMPLEMENT the TOS.
NOTEX1	POP	B
	POP	H		;Backup to start again.
	JMP	RELEXPR 	;Get a non-negated <rel.expr>.
;
;
RELEXPR CALL	ADDEXPR
	RC
	CALL	BREAK
	CALL	TEST
	DB	'<=',0
	JC	RELEXP1
	CALL	ADDEXPR
	RC
	JMP	STKLE
RELEXP1 MVI	A,'<'
	CALL	TESTC
	JC	RELEXP2
	CALL	ADDEXPR
	RC
	JMP	STKLT
RELEXP2 CALL	TEST
	DB	'>=',0
	JC	RELEXP3
	CALL	ADDEXPR
	RC
	JMP	STKGE
RELEXP3 MVI	A,'>'
	CALL	TESTC
	JC	RELEXP4
	CALL	ADDEXPR
	RC
	JMP	STKGT
RELEXP4 MVI	A,'='
	CALL	TESTC
	JC	RELEXP5
	CALL	ADDEXPR
	RC
	JMP	STKEQ
RELEXP5 CALL	TEST
	DB	'!=',0
	JC	SUCCEED
	CALL	ADDEXPR
	RC
	JMP	STKNEQ
;
;
ADDEXPR CALL	MULTEXPR
	RC
	JMP	ADDLIST
;
ADDLIST CALL	BREAK
	MVI	A,'+'
	CALL	TESTC
	JC	ADDL1
	CALL	MULTEXPR
	RC
	CALL	STKADD		;ADD top two elements.
	RC
	JMP	ADDLIST
ADDL1	MVI	A,'-'
	CALL	TESTC
	JC	SUCCEED
	CALL	MULTEXPR
	RC
	CALL	STKSUB		;SUB top from next to top.
	RC
	JMP	ADDLIST
;
;
MULTEXPR CALL	SGDPRIM
	RC
;
MULTLIST CALL	BREAK
	MVI	A,'*'
	CALL	TESTC
	JC	MLTL1
	CALL	SGDPRIM
	RC
	CALL	STKMPY		;MULTIPLY top two items.
	RC
	JMP	MULTLIST
MLTL1	MVI	A,'/'
	CALL	TESTC
	JC	MLTL2
	CALL	SGDPRIM
	RC
	CALL	STKDIV		;DIVIDE next to top by top.
	RC
	JMP	MULTLIST
MLTL2	CALL	TEST
	DB	'MOD',0
	JC	MLTL3
	CALL	SGDPRIM
	RC
	CALL	STKMOD		;MOD next to top by top.
	RC
	JMP	MULTLIST
MLTL3	CALL	TEST
	DB	'SHL',0
	JC	MLTL4
	CALL	SGDPRIM
	RC
	CALL	STKSHL
	RC
	JMP	MULTLIST
MLTL4	CALL	TEST
	DB	'SHR',0
	JC	SUCCEED
	RC
	CALL	SGDPRIM
	RC
	CALL	STKSHR
	RC
	JMP	MULTLIST
;
;
SGDPRIM CALL	BREAK
	MVI	A,'+'
	CALL	TESTC
	JNC	SGDPRIM
	MVI	A,'-'
	CALL	TESTC
	JC	PRIMARY
	MVI	A,'-'
	CALL	TESTC		;Two minuses cancel like one +
	JNC	SGDPRIM		;Can have ----A = A.
	CALL	PRIMARY		;Nope. odd number of '-' signs.
	RC
	CALL	STKNOT		;Negate value of PRIMARY by
	RC
	LXI	D,1		;complementing and incrementing
	CALL	STKPUSH
	RC
	JMP	STKADD
;
;********************************************************
;*		   PRIMARY PARSER			*
;*							*
;*	A primary is the basic operand in an EXPR.	*
;*  It may include: A 16 bit integer valued symbol	*
;*  reference, a numeric constant with optional base	*
;*  specification which evaluates to a 16 bit integer,	*
;*  a $ sign, which has the value of the current	*
;*  location counter, a character string quoted by	*
;*  apostrophes with two or one characters, a percent	*
;*  sign '%' which is the 16 bit checksum of all	*
;*  previous object code bytes, or any EXPR in paren-	*
;*  thesis nested up to 3 levels deep.			*
;*							*
;******************************************************** 
;
PRIMARY CALL	BREAK
	MVI	A,'('		;First, check EXPR in parens.
	CALL	TESTC
	JC	PRM1
	CALL	RECURSE		;CHECK RECURSION DEPTH!!
	RC			;If not too deep, can recurse.
	CALL	EXPR1		;Yes - good, must have expr.
	PUSH	PSW		;EXPR1 doesn't init NESTLVL.
	CALL	UNRECRSE	;Unwind recursion counter
	POP	PSW
	RC
	CALL	BREAK		;Test for matching rgt. paren.
	MVI	A,')'
	CALL	TESTC
	RNC
	LXI	D,0201H		;** UNBALANCED PARENTHESIS **
	STC
	RET
PRM1	MVI	A,'$'
	CALL	TESTC
	JC	PRM110
	PUSH	H		;Save buffer pointer.
	LHLD	LOCATION	;$ gets location counter.
	XCHG
	POP	H
	JMP	STKPUSHR	;Push relocatable value.
;
PRM110	MVI	A,'%'		;% gets running checksum.
	CALL	TESTC
	JC	PRM11
	PUSH	H
	LHLD	CHECKSUM
	XCHG
	POP	H
	JMP	STKPUSH
;
; Check for one of the functions: LEN[ ],NULL[ ],COUNT[ ].
;
PRM11	;
	IF	MACROS
	CALL	TEST
	DB	'LEN[',0	;Space between 'LEN' and left
	JNC	FNLEN		;bracket not allowed.
	CALL	TEST
	DB	'NULL[',0
	JNC	FNNULL
	CALL	TEST
	DB	'COUNT[',0
	JNC	FNCOUNT
	ENDIF			;End MACROS
;
; Check for a SYMBOLIC NAME reference.
;
PRM2	CALL	VAR		;VARIABLE REFERENCE?
	JC	PRM3
	LDA	PASSFLAG	;Yes - OK, what pass we on?
	CPI	1
	JNZ	PRM22		;jmp/ pass two.
	;
	; Got a symbol reference in pass ONE.
	;
	PUSH	D		;Error code will kill DE.
	CALL	SYMFIND		;Has it been seen yet?
	JNC	PRM211		;jmp/ found. Stack value.
	PUSH	H
	LXI	H,0C05H		;Want SYMFIND to ret this code
	CALL	DCMP		;or something worse than
	POP	H		;just not finding it happened.
	JZ	PRM210
	STC
	RET
PRM210	POP	D
	CALL	SYMENTER	;It's pass 1, so enter a
	RC			;symbol of type "UNDEF LABEL"
	MVI	A,6		;Undefined label code.
	CALL	SYMINSERT
	RC
	XRA	A		;Present value just zero
	CALL	SYMINSERT
	RC
	XRA	A
	CALL	SYMINSERT
	RC
	XRA	A		;Reserve space for type.
	CALL	SYMINSERT
	RC
	JMP	PRMZERO 	;Give artificial value of 0
PRM211	POP	D
	LXI	D,0
	CALL	SYMGET
	RC
	CPI	6		;UNDEF LABEL type?
	JNZ	PRM221		;Jmp/ nope, go get value
	;
	;	Give an artificial value of zero and set
	;  BADTERM to say we did.
	;
PRMZERO MVI	A,TRUE		;Set the flag.
	STA	BADTERM
	LXI	D,0		;Fake value of zero.
	JMP	STKPUSHR 	; push as a relocatable.
	;
	; Got a symbol reference in pass TWO.
	;
PRM22	CALL	SYMFIND
	JNC	PRM223
	LXI	D,0E01H 	;## UNDEF LABEL - PASS 2 ##
	RET			;(CARRY by JNC)
PRM223	LXI	D,0		;Zeroth element is type
	CALL	SYMGET
	RC
	CPI	6		;UNDEF-LABEL type?
	JNZ	PRM221
	LXI	D,1101H		;** UNDEFINED SYMBOL REF. **
	CALL	WARN
	JMP	PRMZERO 	;Use value of zero.
;
;  Get the value of the symbol.  Type code already in A.
;OK, check out possible legal types of thing, and get the 
;value if one is found.  Legal are:
;
PRM221	CPI	1		;OPCODE
	JZ	PRM222
	CPI	3		;REGISTER
	JZ	PRM222
	CPI	4		;LABEL
	JZ	PRM222
	CPI	5		;SET-VARIABLE
	JZ	PRM222
	CPI	7		;DDEF-LABEL
	JZ	PRM222
	LXI	D,1102H		;** ILLEGAL SYMBOL TYPE **
	CALL	WARN
	JMP	PRMZERO 	;Give it a zero value.	
	;
PRM222  CALL	SYMVAL		;Get the value.
	PUSH	D
	LXI	D,3
	CALL	SYMGET		;Get type byte
	JC	PRM224
	POP	D
	ORA	A		;Depending on symbol's type
	JZ	STKPUSH		; push either ABS
	JMP	STKPUSHR	; or REL value.
PRM224	POP	PSW
	STC
	RET
;
;  Test for a one or two character string to pack into a
; 16 bit double-word.
;
PRM3	CALL	STRING		;A one or two char string.
	JC	PRM4		;Nope - go try NUMBER.
	CALL	STKPOP		;How big is string?
	PUSH	H
	LXI	H,1
	CALL	DCMP
	POP	H
	JZ	SUCCEED		;jmp/ single char to pack.
PRM32	PUSH	H
	LXI	H,2
	CALL	DCMP
	POP	H
	JZ	PRM31
	PUSH	D
	CALL	STKPOP
	POP	D
	RC
	DCX	D
	PUSH	D
	LXI	D,0203H		;** STRING TOO LONG FOR DW **
	CALL	WARN
	POP	D
	JMP	PRM32
PRM31	CALL	STKPOP
	RC
	MOV	B,E
	CALL	STKPOP
	RC
	MOV	D,E		;Take two DWs on stack, and
	MOV	E,B		;arrange first string char
	JMP	STKPUSH		;in MSB of packed DW.
;
;  Test for a numeric constant.
;
PRM4	MOV	A,D		;STRING should return 0000H
	ORA	E		;code for simple mismatch.
	JNZ	FAIL		;If really bad, FAIL it.
	CALL	NUMBER		;It last possibility: a #.
	RNC
	LXI	D,0204H		;** INVALID OPERAND **
	STC
	RET
;
SUCCEED STC
	CMC
	RET
FAIL	STC
	RET
;
RECURSE LDA	NESTLVL
	INR	A
	STA	NESTLVL
	CPI	6
	CMC
	RNC			;No more than 6 recursions
	LXI	D,0801H		;** EXPRESSION TOO NESTED **
	RET			;Ret. CARRY
;
UNRECRSE LDA	NESTLVL
	ORA	A
	JZ	FAIL		;Can't decrement zero.
	DCR	A
	STA	NESTLVL
	STC
	CMC
	RET
;
;
;********************************************************
;*		FUNCTION EVALUATORS			*
;*							*
;*	Handle the calculation of the LEN[ ],NULL[ ],	*
;*  and COUNT[ ] functions.				*
;*							*
;********************************************************
;
	IF	MACROS
FNLEN	CALL	STKMARK
	PUSH	D
	CALL	EXPRLIST	;Evaluate a string of bytes.
	JC	FNLEN1
	CALL	STKPOP		;Get length of string.
	JC	FNLEN1
	CALL	STKBACK		;Throw away the actual string.
	CALL	STKPUSH		;Leave only the length.
	JC	FNLEN1
	CALL	BREAK
	MVI	A,']'		;Eat right bracket.
	CALL	TESTC
	JNC	FNLEN2
	LXI	D,2301H 	;** Missing bracket **
	JMP	FNLEN1
FNLEN2	POP	D
	CALL	STKUNMARK
	RET			;(NO CARRY by STKUNMARK)
FNLEN1	XCHG
	POP	D
	CALL	STKUNMARK
	XCHG			;Error code in DE.
	STC			;(CARRY)
	RET
;
FNNULL	CALL	PARM
	RC
	LXI	D,0		;FALSE value.
	MOV	A,B		;Is null if B=0.
	ORA	A
	JNZ	FNNULL1
	DCX	D		;Convert to TRUE value (-1).
FNNULL1 CALL	STKPUSH
	RC
FNNULL2 CALL	BREAK
	MVI	A,']'		;All chars that stop the
	CALL	TESTC		;parm parser are taken care
	RNC			;of here.
	MVI	A,';'
	CALL	TESTC
	JNC	FNNULL4
	CALL	COMMA		;Eat possible comma.
	CALL	PARM		;Eat possible parameter.
	MOV	A,C		;Is there any more text?
	ORA	A
	JNZ	FNNULL2		;jmp/ no.
FNNULL4 LXI	D,2301H 	;** Missing brackets **
	STC
	RET
;
FNCOUNT LXI	D,2302H       ;** No COUNT FCN yet - Sorry! **
	STC
	RET
	ENDIF			;End of MACROS.
