	PAGE
;********************************************************
;*							*
;*	     STATEMENT PARSER / EVALUATOR		*
;*							*
;*	The value of a call to the statement parser is	*
;*  either some data on the VSTACK, which should be 	*
;*  DUMPed as object code (from bottom of the VSTACK	*
;*  reading upwards), or it is an alteration in the 	*
;*  state of various flags, such as the ENDFLAG or the	*
;*  SKIP variable.  The evaluation of pseudo-ops 	*
;*  results in these flags and variables changing, 	*
;*  whereas if an instruction opcode is found and	*
;*  evaluated, the result is a list of bytes on the 	*
;*  VSTACK (in the usual format of having TOS be the 	*
;*  byte count).					*
;*	The important memory locations that are		*
;*  modified by the statement parser are:  ENDFLAG,	*
;*  MACDEF,SKIP,WARNING,ERR,IFNEST,IFFAILS,		*
;*  ELSEFLAG,AFIELD,AFLDBLK, and LOCATION.		*
;*	As per all the other parsers, HL -> text to	*
;*  be parsed on input, and C = # text chars.  On	*
;*  exit, DE has error code of fatal errors with CARRY	*
;*  but if no CARRY, then statement was parsed		*
;*  successfully.  In any case, the WARNING word may	*
;*  have a non-zero value, in which case is contains a	*
;*  non-fatal error code, such as:			*
;*    ** UNDEFINED SYMBOL REFERENCE IN EXPRESSION **	*
;*  which is indicated by an 1101H code in WARNING.	*
;*							*
;********************************************************
;
; The meaning of the various flags used by STATEMENT are:
;
;SKIP:		The number of bytes to be emitted as zeroes
;	or to be skipped if a relocating object code format is
;	available.  Set non-zero by the DS or forward-moving
;	ORG pseudo-ops.
;
;WARNING:	If non-zero, contains non-fatal error code.
;	Statement was parsed and produced proper number of
;	bytes of result on VSTACK, but value may be wrong.
;
;ERR:		The error code for any real error that halted
;	parsing of the current line.
;
;GENYET:	Referenced by ORG pseudo-op.  If code has
;	been generated yet, a backwards ORG is illegal (due
;	to strict sequential form required of direct load
;	modules on POLY 8813 system).  GENYET should be set
;
;ENDFLAG:	Set true by the END pseudo-op.
;
;MACDEF:	Set true by the MACRO pseudo-op.
;
;MLISTFLAG:	Set by the MACLIST pseudo-op.  Its value is 
;	used as the initial value of the LISTFLAG within 
;	macro expansions.  LISTFLAG is restored on exit from 
;	a macro expansion.
;
;MACKEY:	Has the symbol table key of the current macro
;	being expanded.  If zero, none being expanded.
;	If STATEMENT finds a macro reference in the opcode
;	field, MACREF is called to build an activation record
;	in memory at MSTG-MSTGEN.
;
;EXPANDP:	Has the pointer to the last macro activation
;	record on the string list stack.  If zero, macros
;	are not stacked on one another.  If MACKEY is zero,
;	EXPANDP is undefined.
;
;LOCATION:	Referenced by EXPR with the $ operand, and
;	used as value of labels being defined.  After the
;	statement has been parsed, LOCATION is incremented
;	by the length of the data on the VSTACK that is to
;	be DUMPed, aded to the value of SKIP.
;
;LABPTR:	Points at the memory location of the text
;	of the last label identifier parsed.  Used to
;	pass the label identifier to the specialized
;	processing statement processing routine, which may
;	use the label for any of it's own purposes, (usually
;	just for defining a new label symbol, but could
;	be a macro name, for example). The LABDEF subroutine
;	will define the pointed at symbol giving it the
;	value passed in in HL.  If LABLEN is zero, no label
;	is to be defined, and LABDEF does nothing.
;
;LABLEN:	The length of the label pointed to by
;	LABPTR.  If zero, LABPTR is not valid, and
;	there was no label in the label field.
;
;OPCPTR:	Similar to LABPTR and LABLEN.
;OPCLEN:
;
;PASSFLAG:	Referenced only.  If PASSFLAG is 1, symbols
;	will be defined and if PASSFLAG is 2, symbols will
;	be referenced and used.  See PRIMARY parser for
;	details.
;
;AFIELD:	Address field value for listing generation.
;	Defaults to LOCATION, but may be altered to
;	show the value given to an EQU or SET variable.
;	The value of AFIELD is intended to be printed to
;	the left of the VSTACK data.
;
;AFLDBLK:	Address field blank.  If AFLDBLK=0FFH, the
;	value of AFIELD should not be printed.  Unless
;	altered by the statement handler, defaults to 0.
;
;PTBASE:	The base of the address and obect code
;	fields on the assembly listing.  Modified by
;	the PRBASE pseudo-op.
;
;LISTFLAG:	This flag is set by the LIST pseudo-op.
;	It has one of four possible states:
;	   0 :  no listing is generated.
;	   2 :  partial listing (only lines generating object
;		code are listed) with the object field only 
;		printed.
;	   3 :  partial listing with both object & source 
;		fields printed.
;    default :  full listing is being generated.
;
;INVISIBLE:	If TRUE, the instruction just assembled is
;	not to be printed.  Changing the LISTFLAG affects the
;	printing of the statement that changed it.  INVISIBLE
;	gives a statement control over whether it itself is
;	printed.  It is a local (one-time) control, as opposed
;	to the global LISTFLAG.
;
;VISIBLE:	Similar to INVISIBLE, but guarantees the
;	current statement prints, rather than guaranteeing
;	that is does not.
;
STATEMENT PUSH	H
	LXI	H,0
	SHLD	SKIP		;Default don't generate zeroes
	SHLD	WARNING		;Clear out warning errors
	SHLD	ERR		; and regular errors.
	CALL	STKINIT 	;Wipe out VSTACK garbage.
	MVI	A,0FFH
	STA	AFLDBLK		;Default address blanked.
	XRA	A
	STA	LABLEN		;No symbol there yet.
	STA	ENDFLAG		;Default not END pseudo-op.
	STA	INVISIBLE	;Default not forced invisible.
	STA	VISIBLE		;Also not forced visible.
	POP	H		;Time to start parse.
	;
	CALL	STMT0
	JC	STMT00		;If bad news, save code.
	CALL	STKPOP		;Good parse. get length of
	JNC	STMT003		;results on stack.
	LXI	D,0		;If nothing at all, pretend 0.
STMT003 CALL	STKPUSH		;Bump LOCATION by that much.
	RC			;Shouldn't ever return.
	MOV	A,D
	ORA	E		;If we are moving the
	JZ	STMT004		;LOCATION counter, we are
	MVI	A,0FFH		;generating code, so set
	STA	GENYET		;the GENYET flag.
STMT004 LHLD	LOCATION
	DAD	D		;Bump LOCATION counter.
	SHLD	LOCATION
	ORA	A
	RET
	;
	;  Parsing failed.  Error code in DE.
	;
STMT00	LXI	H,0C01H 	;** Symbol table full **
	CALL	DCMP
	JNZ	STMT001
	STC			;A real fatal error!
	RET
	;
STMT001 XCHG			;Just very bad news.
	SHLD	ERR		;Save error code
	CALL	STKINIT 	; and wipe out VSTACK.
	LXI	D,0
	JMP	STKPUSH 	;Pretend nothing at all
;
;  This top-level-subroutine actually parses the statement.
;
STMT0	MVI	A,';'		;Is this a comment line?
	CALL	TESTC
	JC	STMT01
	LDA	MACDEF		;If defining a macro, go
	ORA	A		;put it in the prototype.
	JNZ	STMT1131
	MVI	A,0FFH
	STA	AFLDBLK		;Don't print any Address junk.
	XRA	A
	RET
STMT01	LDA	MACDEF
	ORA	A
	JZ	STMT02		;Jmp/ no MACROs being defined.
	CALL	FINDBREAK	;Skip past label
	JMP	STMT1
STMT02	CALL	VAR		;Do we have a label?
	JC	STMT1
	XCHG
	SHLD	LABPTR		;Save label field pointer.
	XCHG
	MOV	A,B
	STA	LABLEN		;and label length pointer.
	MVI	A,':'		;Optional colon after label.
	CALL	TESTC
;
; Now we must interpret the OPCODE field.  The LABEL field
; is handled by whatever handling routine deals with the
; OPCODE field, eg: the pseudo-op handler may use it to define
; a SET variable, if the SET pseudo-op is found.
;
STMT1	CALL	BREAK		;Probably space to opcode fld.
	MVI	A,';'		;Is it a comment?
	CALL	TESTC
	JNC	STMT110
	MOV	A,C		;Or is it a blank line?
	ORA	A		;(it's blank if no more buff.)
	JNZ	STMT11		;No - go interpret opcode fld.
STMT110 LDA	MACDEF		;Are we defining a macro?
	ORA	A
	JNZ	STMT1131	;jmp/ yes: go add to prototype.
	MVI	A,0FFH		;Blank address printing.
	STA	AFLDBLK
	LDA	LABLEN		;Are we going to define a
	ORA	A		;label?
	RZ
	LDA	IFFAILS
	ORA	A
	JNZ	SUCCEED
	XRA	A
	STA	AFLDBLK
	JMP	LABDEF
;
;  Now finally got to opcode itself.  However, we may be
;  skipping text in an IF/ENDIF that failed, or we may be
;  defining a MACRO.  If either one, we look for certain
;  literal opcodes and ignore others altogether:  others
;  are just comments.  Note that in practice, MACDEF and
;  IFFAILS are mutually exclusive.  We assume this here.
;
STMT11	PUSH	H		;But first,
	PUSH	B
	CALL	VAR
	XCHG
	SHLD	OPCPTR		;save the opcode field pointer.
	MOV	A,B		;Also length. (possibly zero).
	STA	OPCLEN
	POP	B
	POP	H
;
; Now check for IF/ENDIF that failed.
;
	LDA	IFFAILS		;Nesting of failed 'IF's.
	ORA	A		;If not zero, must be
	JZ	STMT113		;in middle of failed 'IF'.
	MVI	A,0FFH
	STA	AFLDBLK
	LDA	OPCLEN		;(Opcode length.)
	CPI	5
	JNZ	STMT111
	CALL	TEST		;Yes - middle of 'IF'.
	DB	'ENDIF',0
	JC	STMT111		;No 'ENDIF' - try 'IF'.
	LDA	IFFAILS		;Found 'ENDIF'.
	DCR	A
	STA	IFFAILS		;This one all over.
	JNZ	SUCCEED
	STA	ELSEFLG		;Tricky! reset ELSEFLG if we
	RET			;just returned to active level.
STMT111 LDA	OPCLEN
	CPI	2
	JNZ	STMT112
	CALL	TEST
	DB	'IF',0		;Got to watch nested 'IF's
	JC	STMT112		;Try something else.
	LDA	IFFAILS		;Found another 'IF', but
	INR	A		;keep skipping code even
	STA	IFFAILS		;if it would succeed.
	XRA	A
	RET
STMT112 LDA	IFFAILS
	CPI	1
	JNZ	SUCCEED
	LDA	ELSEFLG
	ORA	A
	JNZ	SUCCEED
	LDA	OPCLEN		;(This test is for prefix
	CPI	6		;problems).
	JNZ	STMT1121
	CALL	TEST
	DB	'ELSEIF',0	;ELSEIFs/ELSEs count only if
	JC	STMT1121	;at level one and haven't
	XRA	A		;succeeded a previous else yet.
	STA	IFFAILS
	JMP	DOIF
STMT1121 LDA	OPCLEN
	CPI	4
	JNZ	SUCCEED
	CALL	TEST
	DB	'ELSE',0
	CMC
	RNC
	XRA	A
	STA	IFFAILS
	LDA	IFNEST		;IFFAILS+IFNEST=constant
	INR	A		;for ELSE/ELSEIF because
	STA	IFNEST		;total nesting constant.
	RET
;
;	Now check for macro definition.
;
STMT113	LDA	MACDEF		;Defining a macro?
	ORA	A		;If not, skip to active
	JZ	STMT114 	;case code.
	MVI	A,0FFH		;Yes.  Blank the Address field
	STA	AFLDBLK		;in listing.
	LDA	OPCLEN		;Opcode length must be
	CPI	4		;4 to match 'ENDM'.
	JNZ	STMT1133
	CALL	TEST		;Also, must look like 'ENDM'.
	DB	'ENDM',0
	JC	STMT1131
	;
	; Found 'ENDM'.
	;
	LDA	MACDEF		;Undo definition depth one
	DCR	A		;level.
	STA	MACDEF
	JNZ	STMT1131
	RET
	;
	; Got 'MACRO' pop nested within a
	; definition?  If so, bump the nesting depth, MACDEF.
	;
STMT1133 CPI	5		;Is OPCLEN=5? if not, can't be
	JNZ	STMT1131	;'MACRO' pop.
	CALL	TEST		;It is right len. - so test
	DB	'MACRO',0		;name.
	JC	STMT1131	;jmp/ nope.
	LDA	MACDEF		;Got a nested 'MACRO', so bump
	INR	A		;nesting level.
	STA	MACDEF		;And fall through.
	;
	; Write the text out to the symbol table definition
	; of the macro.
	;
STMT1131 LDA	PASSFLAG	;Do nothing unless it's pass 1.
	CPI	1
	JNZ	SUCCEED
	LXI	D,LINEBUF	;It is pass 1.
	CALL	DSUB
	MOV	A,L
	ADD	C
	MOV	C,A		;Backup to find linebuf length.
	LXI	H,LINEBUF	;HL,C->original text line.
STMT1132 MOV	A,C
	ORA	A		;Copy until C=0.
	RZ
	MOV	A,M		;Put chars in symtab.
	CALL	SYMINSERT
	RC			;Possible symtab overflow.
	INX	H
	DCR	C
	JMP	STMT1132
;
;	Not skipping code.  Go for real opcode.
;
STMT114 CALL	STMT115 	;Parse statement
	RC			;Ret/ pass along error
	;
	;	Now that the instruction has been parsed
	;  we must either have a comment or be at the end
	;  of the input line.
	;
	CALL	BREAK		;Past blanks.
	JC	SUCCEED 	;Jmp/ end of line, return
	MVI	A,';'
	CALL	TESTC		;Should have comment delimiter
	RNC			;Ret/ yup, it's O.K.
	LXI	D,0204H 	;Operand wasn't completely
	CALL	WARN		; parsed so warn, user.
	ORA	A
	RET
;
;  Parse a statement.
;
STMT115 CALL	INSTR
	JC	STMT12
	JMP	LABDEF		;Jmp/ define possible label.
STMT12	PUSH	H
	LXI	H,0302H		;If non-opcode symbol found by
	CALL	DCMP		;INSTR, we test POPs..
	POP	H
	JZ	STMT2		;Yep - check other kinds...
	STC			;Nope - INSTR blew,
	RET			;so ret. w/ it's error code.
;
STMT2	LXI	D,0		;What is type code?
	CALL	SYMGET		;NOTE: SYMFIND has been
	RC			;called by INSTR.
	CPI	2		;POPCODE?
	JZ	POPREF
	;
	IF	MACROS
	CPI	8		;MACRO reference?
	JZ	MACREF
	ENDIF
	;
	LXI	D,0305H	;** OPCODE FLD: ILLEGAL SYMBOL TYPE **
	STC
	RET
;
; 	A Pseudo-opcode reference.
;
POPREF	LXI	D,1		;Get VALUE of POPCODE (!)
	CALL	SYMGET		;This gets low order byte.
	CPI	EQUCODE
	JZ	DOEQU
	CPI	ORGCODE
	JZ	DOORG
	CPI	IDNTCODE
	JZ	DOIDNT
	CPI	ENDCODE
	JZ	DOEND
	CPI	SETCODE
	JZ	DOSET
	CPI	DSCODE
	JZ	DODS
	CPI	DBCODE
	JZ	DODB
	CPI	DWCODE
	JZ	DODW
	CPI	PRBASECODE
	JZ	DOPRBASE
	CPI	LISTCODE
	JZ	DOLIST
	CPI	IFCODE
	JZ	DOIF
	CPI	ENDIFCODE
	JZ	DOENDIF
	CPI	ELSECODE
	JZ	DOELSE
	CPI	ELSEIFCODE
	JZ	DOELSEIF
	CPI	PAGECODE
	JZ	DOPAGE
	CPI	TITLECODE
	JZ	DOTITLE
	CPI	LOCCODE
	JZ	DOLOC
	CPI	RELOCCODE
	JZ	DORELOC
	;
	IF	REFSDEFS
	CPI	REFSCODE
	JZ	DOREFS
	CPI	REFCODE
	JZ	DOREF
	CPI	DEFSCODE
	JZ	DODEFS
	CPI	DEFCODE
	JZ	DODEF
	ENDIF			;End of REFSDEFS
	;
	IF	MACROS
	CPI	MACROCODE
	JZ	DOMACRO
	CPI	ENDMCODE
	JZ	DOENDM
	CPI	MACLISTCODE
	JZ	DOMACLIST
	ENDIF			;End of MACROS
	;
	LXI	D,0F02H		;## POPCODE VALUE ILLEGAL ##
	STC
	RET
;
;  A MACRO reference found.
;
	IF	MACROS
MACREF	SHLD	PARMPTR		;Save HL,C.
	MOV	A,C
	STA	PARMLEN
	MVI	A,0FFH
	STA	AFLDBLK
	;
	; Save the context on the MSTACK.
	;
	LHLD	EXPANDP
	LXI	D,MSTG
	MVI	B,MSTGEN-MSTG
	CALL	MACPUSH		;Save current macro level.
	RC			;If error, blow away.
	SHLD	EXPANDP		;Saved successfully.
	;
	; Initialize storage.
	;
	LXI	H,0
	SHLD	PARMP		;No parms for this level yet.
	SHLD	LABP		;No label field.
	SHLD	OPCP		;No opcode field.
	SHLD	MACINDEX	;Have read no text yet.
	CALL	SYMKEY		;Get key for current macro.
	XCHG
	SHLD	MACKEY
	LHLD	EXPANDCOUNT	;Unique label generator.
	INX	H		;Count this expansion.
	SHLD	EXPANDCOUNT
	SHLD	EXPANDCNT	;Put it in the local copy.
	LDA	LISTFLAG	;Print the invocation line
	ORA	A
	JZ	MACRF1		;Convert LISTFLAG to boolean.
	MVI	A,0FFH
MACRF1	STA	VISIBLE 	;if LISTFLAG was on.
	CMA
	STA	INVISIBLE
	LDA	MLISTFLAG	;Within a macro, the LISTFLAG
	STA	LISTFLAG	;is defaulted to MACLIST.
	;
	; Bind the parameters.
	; First, the ones from the operand field.
	;
	LHLD	PARMPTR
	LDA	PARMLEN
	MOV	C,A
	CALL	PARM		;Parse a parameter.
	JC	MACRF2		;If none, go try for a label.
	PUSH	H
	LXI	H,0
	CALL	MACPUSH		;Create an entry for that one.
	SHLD	PARMP
	SHLD	PARMPTR		;A temp for construction of
				;backwards list.
	POP	H
	RC			;If macro stack full, CARRY.
MACRF11 MOV	A,C		;Save buffer length.  It
	STA	MACREFTEMP	;should shorten or parm was
	CALL	BREAK		;not found.
	MVI	A,','
	CALL	TESTC		;Optional comma between parms.
	CALL	PARM		;Are there more parms?
	JNC	MACRF13 	;Report errors.
	CALL	WARN		;And go for a label.
	JMP	MACRF2
MACRF13 LDA	MACREFTEMP	;Did buffer shorten?
	CMP	C		;If not, jmp out of infinite
	JZ	MACRF2		;loop.
	PUSH	H		;Save parsing point.
	LXI	H,0		;Each new one is the new end.
	CALL	MACPUSH		;HL->new stack entry.
	JNC	MACRF12
	POP	H
	RET			;(CARRY by JNC)
MACRF12 XCHG
	LHLD	PARMPTR		;Link last one to the new one.
	XCHG
	SHLD	PARMPTR
	XCHG			;HL->old top of stack, DE->new.
	CALL	MACPOINT	;Make the old one point at new.
	POP	H
	JMP	MACRF11
	;
MACRF2	PUSH	H		;Save parse point.
	CALL	MACRF21
	POP	H
	RET
	;
	; Put the (possible) label found on the macro
	; reference on the LABP list.
	;
MACRF21 LDA	LABLEN
	ORA	A
	JZ	MACRF22
	MOV	B,A
	LHLD	LABPTR
	XCHG
	LXI	H,0		;No previous one in this list.
	CALL	MACPUSH
	RC
	SHLD	LABP
	;
	; Put the opcode used in the macro reference on the
	; OPCP list.  (there may be synonyms for a macro).
	;
MACRF22 LDA	OPCLEN
	ORA	A
	RZ	
	MOV	B,A
	LHLD	OPCPTR
	XCHG
	CALL	MACPUSH
	RC
	SHLD	OPCP
	RET
	ENDIF			;End of MACROS.
;	
;
;  The PSEUDO-OP handlers.  (All names start with DO... )
;  See first program segment for the DOREFS...DODEF handlers.
;
DOEQU	CALL	EXPR?
	CALL	STKPOP
	RC
	PUSH	H
	XCHG			;Put value in HL
	LDA	RELFLAG 	; and <rel flag> in A
	CALL	EQUDEF		;Define label.
	POP	H
	RET			;Succeed if EQUDEF does.
;
DOORG	CALL	EXPR
	RC
	CALL	STKPOP
	RC
	PUSH	H
	CALL	DOORG1
	POP	H
	RET
	;
DOORG1	LHLD	LOCATION
	XCHG
	LDA	GENYET		;If no code generated yet,
	ORA	A		;don't worry about negative
	JNZ	DOORG2		;skips.
	SHLD	LOCATION	;Just set LOCATION
	SHLD	LASTREL 	; and last relocated word.
	JMP	LABDEF
	;
DOORG2	PUSH	H
	CALL	DCMP
	JC	DOORG3		;Can't negative skip.  If
	CALL	DSUB		;we try to, we get 1001H.
	SHLD	SKIP		;SKIP=NEW LOC. - OLD LOC.
	POP	H
	SHLD	LOCATION
	JMP	LABDEF		;Define possible label.
DOORG3	POP	H
	CALL	LABDEF?
	LXI	D,1001H    ;** Backwards ORG/ RELOC ignored **
	STC
	RET
;
DOEND	MVI	A,0FFH
	STA	AFLDBLK		;No address junk printed.
	STA	ENDFLAG
	XRA	A
	RET
;
	IF	MACROS
DOENDM	MVI	A,0FFH		;Active case ENDM means we are
	STA	AFLDBLK		;not in macro definition mode.
	LXI	D,0F01H	    ;** ENDM without previous MACRO **
	STC
	RET
	ENDIF			;End of MACROS.
;
	IF	MACROS
DOMACRO MVI	A,0FFH
	STA	AFLDBLK		;Don't put an address field.
	MVI	A,1		;Level one of
	STA	MACDEF		;macro definitions.
	LDA	PASSFLAG
	CPI	1		;Only actually create new
	JNZ	DOMACD1 	;entry if pass 1.
	LHLD	LABPTR
	XCHG
	LDA	LABLEN
	MOV	B,A
	CALL	SYMENTER
	RC			;Blow out if symtab full.
	MVI	A,8		;MACRO type code.
	CALL	SYMINSERT	;(Blow out if symtab full.)
	RC
DOMACD1 MVI	C,0		;"Parse" rest of line.
	ORA	A
	RET
	ENDIF			;End of MACROS.
;
DOSET	CALL	EXPR?		;Don't blow off on errors,
	CALL	STKPOP		;just give a warning.  We have
	RC			;to define a zero value to
	PUSH	H
	XCHG			;Move value to HL.
	LDA	RELFLAG 	; and get <rel flag> in A
	CALL	SETDEF		;Do a SET definition.
	POP	H
	RET
;
DODS	CALL	LABDEF?
	CALL	EXPR?
	CALL	STKPOP
	RC
	XCHG
	SHLD	SKIP
	XCHG
	PUSH	H
	LHLD	LOCATION	;Move LOCATION up that amount.
	DAD	D
	SHLD	LOCATION
	POP	H
	XRA	A
	RET
;
DODB	CALL	LABDEF?
	JMP	EXPRLIST	;Leave data group on VSTACK
;
DODW	CALL	LABDEF?
	LXI	D,0		;Start count at zero.
	PUSH	D
DODW1	CALL	EXPR?
	CALL	ABSADDR 	;Convert into two bytes.
	JC	DODW3		;Jmp/ bad error.
	POP	D		;Got another good EXPR.
	INX	D
	INX	D
	PUSH	D
	CALL	BREAK
	MVI	A,','
	CALL	TESTC		;May have a comma.
	JNC	DODW1		;If so, got another EXPR.
	POP	D
	JMP	STKPUSH		;No - drop in the length.
DODW3	XCHG
	POP	D
	CALL	STKPUSH		;Put admittedly short count
	RC			;on stack.
	XCHG			;And do error return.
	STC
	RET
;
DOPRBASE CALL	LABDEF?
	CALL	EXPR		;Leave base alone if error.
	RC
	CALL	STKPOP
	RC
	MOV	A,D
	ORA	A
	JNZ	DOPR1		;If D <> 0,bad base spec.
	MOV	A,E
	CPI	2		;Base 2
	JC	DOPR1		;Min.
	CPI	26+10+1		;26 letters + 10 digits.
	JNC	DOPR1		;Max.
	XCHG
	SHLD	AFIELD		;Override normal printout.
	SHLD	PTBASE		;Store the new base.
	XCHG
	XRA	A
	RET
DOPR1	XCHG
	SHLD	AFIELD
	LXI	D,0D02H		;** BASE ERROR **
	STC
	RET
;
DOLIST	CALL	EXPR?
	CALL	STKPOP
	RC
	MVI	A,0FFH
	STA	INVISIBLE	;Never list the LIST!
	PUSH	H
	LXI	H,3
	CALL	DCMP
	POP	H
	JNC	DOLIST1 	;Jmp/ EXPR is 0 to 3
	MVI	E,1		; else look like LIST 1
DOLIST1 MOV	A,E
	STA	LISTFLAG
	XRA	A
	RET
;
DOIF	CALL	EXPR?
	CALL	STKPOP
	RC
	XCHG
	SHLD	AFIELD		;Print the expr value.
	XCHG
	XRA	A
	STA	AFLDBLK
	MOV	A,D		;If result is zero, 
	ORA	E		;we have failed 'IF'.
	JZ	DOIF1
	LDA	IFNEST		;Succeeded, so bump nest
	INR	A		;level, but not IFFAILS,
	STA	IFNEST		;the nest level of failures.
	XRA	A
	RET
DOIF1	MVI	A,1		;Start IFFAILS off. (Was zero,
	STA	IFFAILS		;or we wouldn't be here!)
	XRA	A		;(Don't define label.)
	RET
;
DOENDIF MVI	A,0FFH
	STA	AFLDBLK		;Don't print AFIELD
	XRA	A
	STA	ELSEFLG		;Reset for next IF/ENDIF.
	LDA	IFNEST
	ORA	A		;Is there a matching 'IF'?
	JNZ	DOENDIF1
	LXI	D,1501H 	;** ENDIF WITHOUT IF **
	STC			;Failure: CARRY.
	RET
DOENDIF1 DCR	A		;Undo matching 'IF'.
	STA	IFNEST		;(IFFAILS=0, so leave alone.)
	XRA	A		;Success: NO CARRY.
	RET
;
DOELSEIF CALL	EXPR?		;Parse expression.
	CALL	STKPOP
	RC
DOELSE	CALL	DOENDIF		;Active case of ELSE acts like
	RC			;ENDIF/IF 0 pair.
	MVI	A,1		;IFFAILS must have been 0.
	STA	IFFAILS		;Switch from active to
	STA	ELSEFLG		;inactive.  Note further elses
	XRA	A		;ignored by setting ELSEFLG.
	RET
;
DOIDNT	CALL	LABDEF?
	CALL	EXPR?
	CALL	STKPOP
	RC
	XCHG
	SHLD	OLDA		;First opand is Load Addr.
	XCHG
	CALL	COMMA?  
	CALL	EXPR?
	CALL	STKPOP
	RC
	XCHG
	SHLD	OSTA		;Second in Start Addr.
	SHLD	AFIELD		;Print that in Address field.
	XCHG
	XRA	A
	RET
;
	IF	MACROS
DOMACLIST MVI	A,0FFH
	STA	AFLDBLK
	CALL	EXPR?
	CALL	STKPOP
	RC
	PUSH	H
	LXI	H,3
	CALL	DCMP
	POP	H
	JNC	DOMACL1 	;Jmp/ EXPR is 0 to 3
	MVI	E,1		; else look like MACLIST 1
DOMACL1 MOV	A,E
	STA	MLISTFLAG
	XRA	A
	RET
	ENDIF			;End of MACROS.
;
DOPAGE	MVI	A,0FFH
	STA	AFLDBLK
	LDA	PASSFLAG
	CPI	2
	JNZ	SUCCEED
	LDA	LISTFLAG
	ORA	A
	RZ
	LDA	HARDLIST
	ORA	A
	RZ			;Don't put FFs on screen.
	LDA	ERRSONLY	;Don't print if listing only
	ORA	A		;errors.
	JNZ	SUCCEED
	MVI	A,0FFH
	STA	INVISIBLE	;Don't list this line.
	PUSH	B
	PUSH	H
	LXI	H,FFBUF		;Buff containing only
	MVI	C,1		;A form feed.
	CALL	PUTLINE
	POP	H
	POP	B
	XRA	A
	RET
FFBUF	DB	FF
;
;!!!! THIS IS NOT A COMPLETE POP YET !!!!!
;
DOTITLE MVI	A,0FFH
	STA	AFLDBLK		;No address field needed.
	CALL	EXPRLIST	;Title is specified as a
	CC	WARN		;standard EXPRession list.
	PUSH	B
	PUSH	H
	CALL	DOTTL1
	POP	H
	POP	B
	RET
	;
DOTTL1	MVI	A,0FFH
	STA	INVISIBLE	;Don't show this in printout.
	CALL	STKRDINIT	;Prepare to dump title into
	LXI	H,TITLEBUF	;buffer.
	LXI	B,TITLEBLEN
DOTTL2	CALL	STKREAD		;Get title bytes.
	JC	SUCCEED		;When none left, ret.
	MOV	A,B
	ORA	C
	JZ	DOTTL2		;If buf full, eat title chars.
	DCX	B
	MOV	M,E		;Low byte of 16 bit VSTACK el.
	INX	H
	JMP	DOTTL2
;
DOLOC	CALL	EXPR		;Get new location
	RC
	CALL	STKPOP
	RC
	XCHG
	SHLD	LOCATION	;Move LOCATION counter.
	XCHG
	JMP	LABDEF		;Go define possible label.
;
DORELOC LDA	GENYET
	ORA	A
	JZ	DORELC1
	LXI	D,1001H 	;** ORG/ late RELOC ignored **
	STC
	RET
DORELC1 MVI	A,1
	STA	RELOCATE	;Just set RELOC flag
	RET
;
;********************************************************
;*	    INSTR PARSERS UTILITIES			*
;********************************************************
;
; WARN just puts a new warning code (error code) into
; the WARNING cell if it does not already have one.  The
; reason we check first is that frequently an earlier error
; will generate other garbage errors that we don't want.
;	Routines that call WARN when there is some error
; are considered to 'EAT' the thing they parse, because there
; is no recourse if the parse fails: we get an error.
;
WARN	PUSH	PSW
	PUSH	H
	CALL	WARN1
	POP	H
	POP	PSW
	RET
WARN1	LHLD	WARNING
	MOV	A,H
	ORA	L
	RNZ
	XCHG
	SHLD	WARNING
	XCHG
	RET
;
???	CC	WARN		;Indicate possible error,
	STC			;then reset flags because
	CMC			;we don't want to report it
	RET			;twice.
;
ABSADDR CALL	STKPOP		;Expand top VSTACK element
	RC			;into two one-byte elements.
	LDA	RELFLAG
	ORA	A
	JZ	ABSADR1 	;Jmp/ ABS value on TOS.
	MOV	A,D
	MVI	D,0
	CALL	STKPUSHR 	;Push LSB as relocatable.
	RC
	JMP	ABSADR2
ABSADR1 MOV	A,D
	MVI	D,0
	CALL	STKPUSH 	;Push LSB as absolute.
	RC
ABSADR2 MOV	E,A
	MVI	D,0
	JMP	STKPUSH 	; and absolute MSB.
;
;	BYTE? gets a non-optional byte expression from the
; input buffer.  If the expression is relocatable or gives
; a value larger than 8 bits, error results.
;
BYTE?	CALL	BYTE
	JMP	???		;GO set error if one exists.
;
BYTE	CALL	EXPR
	RC
	CALL	STKPOP
	RC
	LDA	RELFLAG
	ORA	A		;Mustn't be relocatable.
	JNZ	BYTE1
	MOV	A,D		;Get high byte.
	MVI	D,0		;Force it zero.
	ORA	A		;Was high byte zero before?
	JZ	STKPUSH		;If so, no warning.
	INR	A		;Is high byte all ones?
	JZ	STKPUSH		;If so, it's OK.
	CALL	STKPUSH		;Else zero the high byte,
	LXI	D,1203H		;** BYTE VALUE > 256 **
	CALL	WARN
	ORA	A
	RET
	;
BYTE1	LXI	D,0205H 	;** Illegal Reloc. Expr. **
	CALL	WARN
	ORA	A
	RET
;
;	COMMA? parses and deletes a comma from front of buffer.
; If not found, it is 0B01 error.  COMMA is a simple scanner,
; just returning CARRY if not found.
;	
COMMA?	CALL	COMMA
	RNC
	LXI	D,0B01H		;** Missing comma **
	JMP	WARN		;WARN preserves CARRY.
COMMA	CALL	BREAK
	MVI	A,','
	JMP	TESTC		;CARRY if no comma.
