	PAGE
;********************************************************
;*			UTILITIES			*
;*							*
;*	All utilities that operate on buffers follow	*
;*  these rules:  INPUT: H-> where to start in buffer.	*
;*  C=max chars left in buffer.  As scanning proceeds,	*
;*  C is decremented, but not beyond zero.  If an 	*
;*  attempt to decrement C below zero is made, the	*
;*  end of the buffer is being bumped, and this is an	*
;*  error that returns a CARRY.  OUT:  H-> char 	*
;*  after last one used by the utility, i.e. the next	*
;*  one to use for the next utility.  A,DE and B are	*
;*  used for specialized parameters.  Functions that	*
;*  have outputs in the form of a new buffer of some	*
;*  sort use DE and C to delimit.  Failure returns	*
;*  CARRY, Success: NOT CARRY.				*
;*							*
;********************************************************
;
; NUM.  Convert number in buffer into binary in D.
; Any problems result in CARRY returned, and error code in
; DE.
; Digit count passed in in C, pointer in HL. 
; NOTE: Digit count must be exact, with no non-digits at
; end.
; IN:	DE=?? 
; OUT:	D=converted number.
;	A=??
;	B=base.
NUM	MOV	A,B
	CPI	26+10		;26 letters + 10 digits.
	JC	NUM7
NUM0	LXI	D,0D02H		;** BASE OUT OF RANGE **
	STC
	RET
NUM7	CPI	2
	JC	NUM0
	LXI	D,0		;Start converted number at 0.
NUM5	CALL	DIGIT
	JC	NUM1
	SUI	'0'
	JMP	NUM2
NUM1	CALL	LETTER
	JC	NUM3
	CALL	Fold		;Make LC into UC.
	SUI	'A'-10		;'A'-('A'-10)=10
NUM2	CMP	B		;Is digit less than base?
	JC	NUM4
	LXI	D,0D01H		;** ILLEGAL DIGIT IN NUMBER **
	STC			;CARRY.
	RET
NUM4	PUSH	H		;Got to convert a valid digit.
	PUSH	PSW		;Save digit value.
	MOV	L,B		;BIN=BIN*BASE.
	MVI	H,0
	CALL	DMPY		;DE=DE*HL.
	JNC	NUM8		;If no overflow, continue.
	POP	PSW
	POP	H
	LXI	D,0D04H		;** NUMBER TOO LARGE **
	RET			;(CARRY by JNC)
NUM8	POP	PSW
	MOV	L,A		;BIN=BIN+DIGIT.
	MVI	H,0
	DAD	D
	XCHG
	POP	H
	JMP	NUM5		;Keep converting digits.
NUM3	MOV	A,C		;Bad digit? could be end buff.
	ORA	A
	JZ	SUCCEED		;Yep, buffer empty - done.
	LXI	D,0D01H		;** ILLEGAL DIGIT IN NUMBER **
	STC
	RET
;
; @NUM.  Convert binary in DE into number in buffer.
; If base in B is bad, we get 0D02H ** BASE OUT OF RANGE **.
; Print field size desired is in C.  All of
; the print field area is used, and is padded on left with
; the character in A.  (If A=0, you can simulate no padding.)
; If print field is exceeded, we return a 0D03H 
; ** PRINT FIELD ERROR ** code.
; The number is converted into characters in the HL,C buffer,
; with the usual shortening of the buffer by the length of
; the printed number.  The number of digits the number 
; required exclusive of the padding is returned in A.
; This is handy for finding the field size required for
; a given maximum value with a given base.
; Of course, if max val is 256 or 65535, PFCALC gives proper
; print field size.
;
@NUM	STA	PSWSAVE
	SHLD	HSAVE
	XCHG
	SHLD	DSAVE
	MOV	H,B
	MOV	L,C
	SHLD	BSAVE
	STA	PSWSAVE
	XRA	A
	STA	PSWSAVE+1
	CALL	@NUM1		;Do the conversion.
	LDA	PSWSAVE+1	;Has # significant digits.
	LHLD	BSAVE
	MOV	C,L
	MOV	B,H
	LHLD	DSAVE
	XCHG
	LHLD	HSAVE		;(H changes during conversion)
	RET
;
@NUM1	LDA	BSAVE+1		;Is base out of range?
	CPI	26+10+1
	CMC
	JNC	@NUM11
@NUM12	LXI	H,0D02H		;** Base out of range **
	SHLD	DSAVE
	RET			;(CARRY by JNC)
@NUM11	CPI	2
	JC	@NUM12
	LHLD	HSAVE
	MVI	B,0
	DAD	B		;Find end of buffer.
	SHLD	HSAVE		;Leave HL bumped over number.
	DCX	H		;Move back to first spot
	SHLD	PCSAVE		;inside buffer for start.
	LHLD	DSAVE		;Get binary to convert
	MOV	B,H
	MOV	C,L		;It stays in BC during loop.
	MOV	A,B		;Is it initially zero?
	ORA	C
	JNZ	@NUM2		;If not, go ahead and convert.
	MVI	A,'0'		;Else print a zero so the
	JMP	@NUM4		;number won't be blank.
@NUM2	MOV	A,C
	ORA	B		;Is it zero yet?
	JNZ	@NUM5		;If so, go padding.
	LDA	PSWSAVE		;Get pad char.
	JMP	@NUM4		;Go pad on left.
@NUM5	LDA	PSWSAVE+1
	INR	A		;Count that significant digit.
	STA	PSWSAVE+1
	LDA	BSAVE+1		;Get base back.
	MOV	E,A
	MVI	D,0
	CALL	DDIV		;BC = BC/DE, DE = BC MOD DE
	MOV	A,E		;Remainder is MSB so far.
	CPI	10
	JNC	@NUM3
	ADI	'0'		;0+'0'='0'
	JMP	@NUM4		;Print the digit.
@NUM3	ADI	'A'-10		;10+('A'-10)='A'
				;Print the letter.
	;
	; Got char in A to stack in over others at
	; back of buffer.
	;
@NUM4	LHLD	PCSAVE
	PUSH	PSW
	LDA	BSAVE		;Is buffer empty yet?
	ORA	A
	JZ	@NUM6		;Yes - all done.
	DCR	A
	STA	BSAVE
	POP	PSW
	MOV	M,A		;Put digit or letter in buff.
	DCX	H		;Putting stuff in backwards.
	SHLD	PCSAVE
	JMP	@NUM2		;Keep getting digits.
;
; Out of print field.  If BC=0, it is OK.
;
@NUM6	POP	PSW		;From the DCMP in @NUM4.
	MOV	A,B
	ORA	C		;Did whole number convert?
	JNZ	@NUM7
	XRA	A		;Yes - good. ret & restore.
	RET
@NUM7	LXI	H,0D03H		;** PRINT FIELD ERROR **
	SHLD	DSAVE		;Tricky error return.
	STC
	RET
;
; The print field calculator.  PFCALC
; takes the base in B.  If the base
; is acceptable, NOT CARRY is returned with the 
; 16 bit print field in A and the 8 bit print field
; in B.  Otherwise, CARRY is returned.
;
PFCALC	CPI	26+10+1		;26 Chars +10 digits.
	CMC
	RC
	CPI	2		;base must not be 0 or 1.
	RC
	PUSH	H
	LXI	H,PFTABLE
PFCLC1	CMP	M	
	JZ	PFCLC2
	JC	PFCLC2
	INX	H
	INX	H
	INX	H
	JMP	PFCLC1
PFCLC2	INX	H
	MOV	A,M		;16 bit print field.
	INX	H
	MOV	B,M		;8 bit print field.
	POP	H
	STC
	CMC
	RET
PFTABLE	DB	2,16,8		;For base 2, 16 or 8 bits.
	DB	3,11,6
	DB	4,8,4
	DB	6,7,4		;Note we skipped 5...
	DB	9,6,3
	DB	15,5,3
	DB	40,4,2		;40 is > 36, so can't oflow.
;
; COPY moves BC bytes from *DE to *HL.
;
COPY	MOV	A,B
	ORA	C
	RZ
	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	DCX	B
	JMP	COPY
;
; FILL fills BC bytes starting at *HL with A.
;
FILL	PUSH	D		;Preserve DE.
	MOV	D,A		;Stash A in D.
FILL1	MOV	A,B
	ORA	C
	JZ	FILL2		;Jmp/ done.
	MOV	M,D
	INX	H
	DCX	B
	JMP	FILL1
FILL2	MOV	A,D		;Restore A.
	POP	D
	RET
;
; SWAP copies the contents of one area of memory into another
;  area of memory and vice versa.  The two pointers are in
;  HL, DE and the length is in BC.  Mashes A.
;
SWAP	MOV	A,M
	PUSH	PSW
	LDAX	D
	MOV	M,A
	POP	PSW
	STAX	D
	INX	H
	INX	D
	DCX	B
	MOV	A,B
	ORA	C
	JNZ	SWAP
	RET
;
; TEST.  Check to see if front of buffer has a given
; literal string.  The string follows call to TEST,
; delimited by 0 byte.  If not found, return CARRY
; and don't move, else return NOT CARRY and move H
; pointer to after word, and decrement C by length of word.
; In other words, if the string is found, it is deleted
; from the buffer, and ret. NO CY.
;
TEST	SHLD	HSAVE		;May need to backup.
	XCHG			;DE pushed, buff ptr to DE.
	XTHL			;RETADD put in HL.
	PUSH	PSW
	MOV	A,C		;May need to backup.
	STA	PSWSAVE	
	XRA	A		;Reet test flag.
	STA	TESTFLAG
	CALL	TEST1		;Do the test.
	LDA	TESTFLAG	;Did it match?
	ORA	A
	JZ	TEST2		;If still zero, it matched.
	LDA	PSWSAVE		;No match - have to backup
	MOV	C,A		;Backup buff length.
	POP	PSW
	XTHL			;DE clean, RETADD pushed.
	XCHG			;Buff ptr to HL.
	LHLD	HSAVE		;Backup buff ptr.
	STC			;Fail - CARRY.
	RET
;
TEST2	POP	PSW		;Clean A.
	XTHL			;DE clean, RETADD pushed.
	XCHG			;Buff pointer back into HL.
	STC
	CMC
	RET
;
TEST1	MOV	A,M		;Look at match char.
	ORA	A		;Found the zero yet?
	JZ	TEST7		;Yes - done scanning.
	MOV	A,C		;The test. first check buff
	ORA	A		;length.  If zero, no match.
	JNZ	TEST8
	MVI	A,0FFH		;No more buffer, so error.
	STA	TESTFLAG
	INX	H		;Bump match ptr only.
	JMP	TEST1		;Keep looking for zero.
TEST8	LDAX	D
	CMP	M		;Are they the same?
	JZ	TEST5
	MVI	A,0FFH		;No - set flag and continue.
	STA	TESTFLAG
TEST5	INX	D		;Bump buffer pointer,
	DCR	C		;and buffer count.
	INX	H		;Move RETADD (match ptr).
	JMP	TEST1		;Keep scanning.
TEST7	INX	H		;Bump over the zero.
	RET			;Done scanning.
;
; TESTC.  Like TEST, but checks only one character, which
; is passed in A.
;
TESTC	PUSH	PSW
	MOV	A,C
	ORA	A
	JZ	TSTC1
	POP	PSW
	CMP	M
	JNZ	FAIL
	INX	H
	DCR	C
	STC
	CMC
	RET
TSTC1	POP	PSW		;No buffer left.
	STC
	RET
;
; COMPARE checks two strings for equality or evaluates
; their lexical relationship (alphabetical values).
; HL->string A, C=length string A, DE->string B,
; B=length string B.  If string A < string B, ret CARRY.
; If strings are equal, with same length, ret ZERO.
; If strings are equal, but of different lengths,
; the shorter is taken as the lexically 'greater', or
; first in the ordering.  Null strings are equal to
; themselves.
;
COMPARE	MOV	A,B		;Are they equal lengths?
	ORA	C		;If so, ret ZERO.
	RZ
	MOV	A,C		;If A string shorter, CARRY.
	ORA	A
	JZ	COMP1
	MOV	A,B		;If B string shorter, NO 
	ORA	A		;CARRY.
	JZ	COMP2
	LDAX	D		;Got to compare chars.
	CMP	M		;If A string char <> B
	CMC			;string char,  ret. with
	RNZ			;the result in CY.
	INX	H		;Got to check further.
	INX	D
	DCR	C
	DCR	B
	JMP	COMPARE		;Keep going.
COMP1	SBI	1		;A was=0 :set CY, reset Z.
	RET			;Note: DCR A doesn't work!
COMP2	ADI	1		;A was=0 :reset both CY and Z.
	RET			;Note: INR A doesn't work!
;
; VAR.  Move past valid variable name in text.  If
; found, H is moved and C is decremented normally,
; but return D-> start of variable name, B=length
; of variable name.  Does not skip leading 
; blanks.
;
VAR	MVI	B,0		;Start with VAR buf empty.
	MOV	D,H		;And pointing at same place
	MOV	E,L		;as HL (input) buffer.
	CALL	LETTER		;Must start with letter,
	JNC	VAR1
	MVI	A,'_'		;an underscore,
	CALL	TESTC
	JNC	VAR1
	MVI	A,'@'		;an at sign,
	CALL	TESTC
	JNC	VAR1
	MVI	A,'?'		;or a question mark.
	CALL	TESTC
	RC
VAR1	INR	B
	CALL	VARCHAR 	;Got a valid VAR, so find end.
	JNC	VAR1
	CMC			;Found end.
	RET			;(Not CARRY by JNC,CMC)
;
; VARCHAR checks for a character that is legal inside
; a VAR.
;
VARCHAR CALL	LETTER
	RNC
	CALL	DIGIT
	RNC
	MVI	A,'_'
	CALL	TESTC
	RNC
	MVI	A,'?'
	CALL	TESTC
	RNC
	MVI	A,'@'
	JMP	TESTC
;
; BREAK.  Delete SPACE, TAB and CRs from front of buffer.
; Effectively, we skip over blank area by bumping HL
; and dec'ing C.  Sets CARRY iff end of line is hit.
;
BREAK	MOV	A,C
	ORA	A
	JZ	FAIL
	MOV	A,M
	INX	H
	DCR	C
	CPI	TAB		;Most likely a TAB.
	JZ	BREAK
	CPI	' '		;then possibily space.
	JZ	BREAK
	CPI	CR		;Or carriage return.
	JZ	BREAK
	CPI	LF		;Last chance - line feed.
	JZ	BREAK
	DCX	H
	INR	C
	STC			;Not one of MY friends, so
	CMC			;leave HL pointing at it.
	RET			;Not CARRY.
;
; DIGIT. Test to see if the first char in buffer
; is a digit 0-9.  If the buffer is empty, of course,
; return CARRY.  If the char is not a digit, ret CARRY
; also.  Else, ret. NOT CARRY.
; The digit, if found, is returned in A.
;
DIGIT	MOV	A,C
	ORA	A
	JZ	FAIL
	MOV	A,M
	CPI	'0'		;Less than '0' non-digit.
	RC			;If less, ret CARRY.
	INX	H		;Prepare for match.
	DCR	C
	CPI	'9'+1		;Less than or eq to '9'?
	CMC
	RNC			;If less, got a digit.
	DCX	H
	INR	C
	STC
	RET
;
;LETTER. Test to see if the first char in buffer
;is a letter A-Z,a-z.  Rest of rules like DIGIT.
;The letter, if found, is returned in A.
;
LETTER	MOV	A,C
	ORA	A
	JZ	FAIL
	MOV	A,M
	CPI	'A'
	RC			; char < cap A not letter.
	CPI	'Z'+1
	JC	LETR1		;But char <= cap Z is letter.
	CPI	'a'
	RC
	CPI	'z'+1
	JC	LETR1
	STC			;Not letter - No skip, CARRY.
	RET
LETR1	INX	H		;Found a letter, so skip it.
	DCR	C
	STC			;No CARRY.
	CMC
	RET
;
; FINDBREAK.  Find a break as defined above.  If we find one
; we jump to break to eat possible successive blanks.
;
FINDBREAK MOV	A,C
	ORA	A
	JZ	FAIL
	MOV	A,M
	INX	H
	DCR	C
	CPI	TAB		;Most likely a TAB.
	JZ	BREAK		;Jmp/ found one, get others
	CPI	' '		;then possibily space.
	JZ	BREAK
	CPI	CR		;Or carriage return.
	JZ	BREAK
	CPI	LF		;Last chance - line feed.
	JZ	BREAK
	JMP	FINDBREAK	;Didn't find a blank, spin.
;
;	**** MATHEMATICAL UTILITIES ****
;
;	DMPY.  A 16 bit integer multiply.
;  Normally, overflow produces 102H error code, but here
;  it has been temporarily squashed - see semicolons before
;  JC DMPY3...
;
DMPY	PUSH	PSW		;Be clean.
	PUSH	B
	PUSH	H
	MOV	B,H
	MOV	C,L
	LXI	H,0
	MVI	A,16
DMPY1	DAD	H
;	JC	DMPY3
	PUSH	PSW		;Save bit count.
	MOV	A,C
	RAL
	MOV	C,A
	MOV	A,B
	RAL
	MOV	B,A		;Top bit of B determines add
	JNC	DMPY2		;or not add.
	DAD	D		;If top bit 1, add D.
DMPY2	POP	PSW		;Restore bit count
	DCR	A
	JNZ	DMPY1		;Repeat 16 times.
	XCHG
	POP	H
	POP	B		;Now HL=HL*DE
	POP	PSW
	STC
	CMC
	RET			;(Not CARRY by JNZ)
;DMPY3	LXI	D,0102H		;$$ MULTIPLY OVERFLOW $$
;	POP	H
;	POP	B
;	POP	PSW
;	STC
;	RET
;
;
; DDIV is a 16 bit integer division.  Quotient returned
; in BC, remainder in DE.
; Division by zero results in return carry, with DE=0101H code.
; OUT	BC=BC/DE
;	DE= BC MOD DE
;
DDIV	PUSH	H		;Save HL.
	MOV	A,E
	ORA	D		;Check for zero divide.
	JZ	DDIV4
	MOV	A,D		;negate divisor.
	CMA
	MOV	D,A
	MOV	A,E
	CMA
	MOV	E,A
	INX	D		;DE=-DE
	LXI	H,0		;Initial remainder.
	MVI	A,17		;Initialize loop counter.
DDIV0	PUSH	H		;Save remainder.
	DAD	D		;Subtract divisor.
	JNC	DDIV1		;If underflow, bring back
	XTHL			; remainder.
DDIV1	POP	H		;else bring back difference.
	PUSH	PSW		;Save loop counter.
	MOV	A,C		;4 register left shift.
	RAL
	MOV	C,A
	MOV	A,B
	RAL
	MOV	B,A
	MOV	A,L
	RAL
	MOV	L,A
	MOV	A,H
	RAL
	MOV	H,A
	POP	PSW		;Restore loop counter.
	DCR	A		;count on it.
	JNZ	DDIV0		;keep going till count zeroes.
	;
	; Post divide cleanup.
	; Shift remainder right and return in DE.
	;
	ORA	A
	MOV	A,H
	RAR
	MOV	D,A
	MOV	A,L
	RAR
	MOV	E,A		;DE=remainder=(BC MOD DE)
	POP	H		;BC=quotient=(BC/DE).
	ORA	A		;No error - NO CARRY.
	RET
DDIV4	POP	H
	LXI	D,0101H 	;Zero divide error code.
	STC
	RET			;CARRY means error.
;
;	DSUB.  16 bit integer subtraction.
;
DSUB	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET
;
;	DCMP.  Compare HL to DE by subtracting DE.
;  The CARRY is set if DE>HL, and ZERO is set if DE=HL.
;
DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET
