	PAGE
;********************************************************
;*		VALUE STACK FUNCTIONS			*
;*							*
;*	The VSTACK is used to accumulate the value of	*
;*  an expression or instruction and its operands as	*
;*  it is being evaluated.  All the VSTACK functions	*
;*  start with the letters STK.  They all save HL and 	*
;*  BC, and return CARRY if an error occurred.  Most	*
;*  frequent errors are stack underflow and overflow,	*
;*  which are 0502H and 0501H codes respectively.	*
;*  DE is the standard 16 bit value parameter passing	*
;*  cell.  The VSTACK is 'marked' by STKMARK and 'un-	*
;*  marked' by STKUNMARK.  After marking, STKPOP un-	*
;*  derflows when the stack is unwound to the same	*
;*  point as it was when the mark occurred.  After	*
;*  unmarking, the stack may be further popped.		*
;*  The function  STKBACK pops the stack down to	*
;*  the mark.  Note this scheme is not a true marking	*
;*  system, since the marking does not recurse.		*
;*	STKREAD and STKRDINIT are for dumping values	*
;*  from the bottom of the stack up.			*
;*							*
;********************************************************
;
; Last updated:
;	2/8/80		RTM/JDW >= and <= are backwards!
;
STKINIT LXI	H,VSTACK	;Initialize VSTACK and
	SHLD	VSTKPTR		;the VSTACK mark pointer.
	SHLD	VSTKMARK
	SHLD	VSTKREAD
	ORA	A		;Clear CARRY.
	RET
;
STKBACK PUSH	H		;Pop stack back to current
	LHLD	VSTKMARK	;mark position.
	SHLD	VSTKPTR
	POP	H
	ORA	A		;Clear CARRY.
	RET
;
STKMARK PUSH	H		;Put mark at current
	LHLD	VSTKMARK	;(Old marker ret in DE).
	XCHG
	LHLD	VSTKPTR		;stack position.  Can't
	SHLD	VSTKMARK	;pop beyond the mark.
	POP	H
	ORA	A		;Clear CARRY.
	RET
;
STKUNMARK XCHG			;Put marker back to spot
	SHLD	VSTKMARK	;pointed to by DE.
	XCHG
	ORA	A		;Clear CARRY.
	RET
;
STKRDINIT PUSH	H
	LXI	H,VSTACK	;Start reading at bottom.
	SHLD	VSTKREAD
	POP	H
	ORA	A		;Clear CARRY.
	RET
;
STKREAD PUSH	H		;We get values from VSTACK
	PUSH	PSW		;until top found.
	LHLD	VSTKPTR
	LXI	D,-4		;Protecion plus entry length
	DAD	D		;(We warn with CARRY on
	XCHG			;next-to-last call.)
	LHLD	VSTKREAD
	CALL	DCMP
	JNC	STKRD1
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	MOV	A,M
	STA	RELFLAG
	INX	H
	SHLD	VSTKREAD
	POP	PSW
	POP	H
	ORA	A		;Clear CARRY.
	RET
STKRD1	POP	PSW
	POP	H
	STC
	RET
;
;STKPUSHR.  Used for recursive calls within the parser.
;Puts current value of a relocatable expression (DE) on
;stack called VSTACK (value stack).  It checks the depth
;of the stack, returning CARRY and 0501H code if too deep.
;
STKPUSHR PUSH	PSW		;Preserve all regs.
	LDA	RELOCATE
	JMP	STKPS1		;Push possible relocatable
;
;STKPUSH is like STKPUSHR accept it pushes an absolute
;value (constant) onto the VSTACK instead of a relocatable
;
STKPUSH PUSH	PSW		;Preserve all regs.
	XRA	A
	;
STKPS1	STA	RELFLAG 	;Set global.
	PUSH	H		;Must preserve HL,DE.
	PUSH	D
	LHLD	VSTKPTR		;Get stack pointer.
	LXI	D,VSTKTOP-3	;Compare w/ end of stack area.
	CALL	DCMP		;(VSTKTOP-VSTKPTR) > 3?
	JC	STKPS2		;Yes - got room to push.
	POP	D
	POP	H
	POP	PSW
	LXI	D,0501H		;$$ VSTACK OVERFLOW $$
	STC
	RET
	;
STKPS2	POP	D		;HL=VSTKPTR, DE=value.
	MOV	M,E		;Push value.
	INX	H
	MOV	M,D
	INX	H
	LDA	RELFLAG 	;Get relocation tag.
	MOV	M,A		; push entry type.
	INX	H
	SHLD	VSTKPTR 	;Update stack pointer.
	POP	H
	POP	PSW
	ORA	A		;Clear CARRY.
	RET
;
;STKPOP.  Complement to STKPUSH.  Checks for stack underflow,
;if so, returns CARRY and 0502H code.  Value returned in DE.
;Underflow means the stack pointer equals the marker pointer
;when STKPOP is called.
;
STKPOP	PUSH	PSW		;Preserve all regs.
	PUSH	H
	LHLD	VSTKMARK
	XCHG
	INX	D
	INX	D
	INX	D
	LHLD	VSTKPTR
	CALL	DCMP		;VSTKMARK+3 > VSTKPTR?
	JNC	STKPP1		;Nope - good, we got room.
	LXI	D,0502H		;## VSTACK UNDERFLOW ##
	POP	H
	POP	PSW
	STC
	RET
STKPP1	DCX	H
	MOV	A,M
	STA	RELFLAG 	;Set relocation flag.
	DCX	H
	MOV	D,M
	DCX	H
	MOV	E,M		;Value in DE.
	SHLD	VSTKPTR		;Update stack pointer.
	POP	H
	POP	PSW
	ORA	A		;Clear CARRY.
	RET
;
;	**** STACK MATHEMATICS FUNCTIONS ****
;
;	These functions do math on the top two (sometimes)
; only the top one) elements of the VSTACK.  They affect
; only A and DE.
;
;
;	GETARGS gets the arguments off the top of the VSTACK.
; If there are not enough elements on the stack, GETARGS will 
; return to the caller of the stack function by popping stack 
; and returning with CARRY set.
;	ARG1(NTOS) returned in DE.  ARG2(TOS) returned in HL.
;
GETARGS SHLD	HSAVE		;Save both HL and BC
	MOV	H,B		; under rocks.
	MOV	L,C
	SHLD	BSAVE
	CALL	GETABS		;Pop absolute.
	JC	GETA1		;Jmp/ not there so go away.
	XCHG			;DE=ARG1(NTOS).
	;
	CALL	GETABS		;Get second arg.
	JC	GETA1		;Jmp/ not there, leave.
	XCHG			;HL=ARG2(TOS)
	RET
;
;	Get an absolute argument from the top of the VSTACK
; into DE.
;
GETABS	CALL	STKPOP		;Pop off value into DE.
	RC			;Ret/ pass error along.
	LDA	RELFLAG
	ORA	A		;Must be ABS.
	RZ			;Yup/ return NO CARRY
	PUSH	D
	LXI	D,0205H 	;** Illegal Reloc. Expr. **
	CALL	WARN		;Warn user.
	POP	D
	ORA	A		;NO CARRY.
	RET
;
;	GETRARGS is the same as GETARGS accept relocatable 
; arguments are permitted.
;	ARG1(NTOS) returned in B,DE.  ARG2(TOS) returned
; in C,HL.
;
GETRARGS SHLD	HSAVE		;Save both HL and BC
	MOV	H,B		; under rocks.
	MOV	L,C
	SHLD	BSAVE
	CALL	STKPOP		;Get first arg.
	JC	GETA1		;Jmp/ not there so go away.
	XCHG			;ARG1(NTOS) in DE
	LDA	RELFLAG
	MOV	B,A		;Rel flag in B.
	;
	CALL	STKPOP		;Get second arg.
	JC	GETA1		;Jmp/ not there, leave.
	XCHG			;HL=ARG2(TOS)
	LDA	RELFLAG
	MOV	C,A		;Rel flag in C.
	RET
	;
	;	Not enough elements on VSTACK.  Pop off stack
	; and directly to the caller of the stack function!
	;
GETA1	POP	H		;Throw away retadd!
	LHLD	BSAVE
	MOV	B,H
	MOV	C,L		;Restore BC.
	LHLD	HSAVE		; and HL.
	STC			;Error code still in DE.
	RET			;Return CARRY.
;
;	PUTARG puts back all the contents of BC and HL by
; looking under the rocks set by GETARGS and pushes DE down 
; the VSTACK.  If PUTARG is called with CARRY from inside 
; the stack function routine, the result is CARRY on return 
; from PUTARG, and the error code, if any, is still in DE.
;
PUTARG	PUSH	D		;Save possible error code.
	XCHG			;STK fcn results -> DE.
	LHLD	BSAVE
	MOV	B,H
	MOV	C,L
	LHLD	HSAVE
	JC	PUTA1		;If STK fcn errs, pass CARRY.
	INX	SP		;Throw DE - not an error code.
	INX	SP
	JMP	STKPUSH
PUTA1	LXI	D,0		;Got to push something so
	CALL	STKPUSH		;VSTACK is predictable.
	POP	D		;This STKPUSH cannot oflow,
	STC			;so leave STK fcn error code
	RET			;in DE.
;
;	PUTRARG is the same as PUTARG accept it pushes the
; result as a relocatable value instead of absolute.
;
PUTRARG PUSH	D		;Save possible error code.
	XCHG			;STK fcn results -> DE.
	LHLD	BSAVE
	MOV	B,H
	MOV	C,L
	LHLD	HSAVE
	JC	PUTRA1		;If STK fcn errs, pass CARRY.
	INX	SP		;Throw DE - not an error code.
	INX	SP
	JMP	STKPUSHR
PUTRA1	LXI	D,0		;Got to push something so
	CALL	STKPUSHR	;VSTACK is predictable.
	POP	D		;This STKPUSH cannot oflow,
	STC			;so leave STK fcn error code
	RET			;in DE.
;
;	The stack math functions themselves.
;
STKADD	CALL	GETRARGS	;Calculate ARG1+ARG2
	DAD	D		;HL=HL+DE
	MOV	A,B
	XRA	C		;Compare types (and NC).
	JNZ	PUTRARG 	;Jmp/ diff, so push REL.
	ORA	A		;Were both ABS?
	JZ	PUTARG		;Jmp/ yup, push ABS.
	;
	LXI	D,0205H 	;** Illegal Reloc. Expr. **
	CALL	WARN		;Give warning.
	JMP	PUTARG		;And push absolute
;
STKSUB	CALL	GETRARGS	;Calculate ARG1-ARG2
	CALL	DSUB		;HL=HL-DE
	MOV	A,C
	ORA	A
	JZ	STKSUB1 	;Jmp/ ARG1 absolute
	MOV	A,B
	ORA	A		;ARG2 absolute?
	JZ	PUTRARG 	;Jmp/ REL-ABS, push REL
	JMP	PUTARG		;else REL-REL, push ABS
	;
STKSUB1 MOV	A,B
	ORA	A		;ARG2 absolute?
	JZ	PUTARG		;Jmp/ ABS-ABS, push ABS
	;
	LXI	D,0205H 	;** Illegal Reloc. Expr. **
	CALL	WARN		;Give warning.
	JMP	PUTARG		;And push absolute
;
STKMOD	CALL	GETARGS
	MOV	B,H		;DDIV wants dividend in BC,
	MOV	C,L		;and leaves remainder in DE.
	CALL	DDIV		;DE=BC MOD DE
	MOV	H,D
	MOV	L,E
	JMP	PUTARG		;Pass CARRY to PUTARG.
;
STKSHL	CALL	GETARGS		;Shift ARG1 left ARG2 times
STKSHL1 MOV	A,D		;putting zeroes in LSB.
	ORA	E
	JZ	PUTARG		;stop when DE goes to zero.
	MOV	A,L
	RAL
	MOV	L,A
	MOV	A,H
	RAL
	MOV	H,A
	DCX	D
	JMP	STKSHL1		;Keep shifting.
;
STKSHR	CALL	GETARGS		;Shift ARG1 right ARG2 times.
STKSHR1 MOV	A,D		; putting in zeros at top.
	ORA	E
	JZ	PUTARG		;stop when DE goes to zero.
	MOV	A,H
	RAR
	MOV	H,A
	MOV	A,L
	RAR
	MOV	L,A
	DCX	D
	JMP	STKSHR1
;
STKMPY	CALL	GETARGS
	CALL	DMPY		;DE=HL*DE
	XCHG
	JMP	PUTARG
;
STKDIV	CALL	GETARGS		;ARG1/ARG2
	MOV	B,H
	MOV	C,L
	CALL	DDIV		;BC=BC/DE
	MOV	H,B
	MOV	L,C
	JMP	PUTARG
;
STKNOT	CALL	GETABS		;Pop off ABS.
	RC
	MOV	A,D
	CMA
	MOV	D,A
	MOV	A,E
	CMA
	MOV	E,A
	JMP	STKPUSH 	;Push back ABS.
;
STKNEG	CALL	GETABS
	RC
	MOV	A,D
	CMA
	MOV	D,A
	MOV	A,E
	CMA
	MOV	E,A
	INX	D		;Two's complement
	JMP	STKPUSH
;
STKOR	CALL	GETARGS
	MOV	A,D
	ORA	H
	MOV	H,A
	MOV	A,E
	ORA	L
	MOV	L,A
	STC
	CMC
	JMP	PUTARG
;
STKXOR	CALL	GETARGS
	MOV	A,D
	XRA	H
	MOV	H,A
	MOV	A,E
	XRA	L
	MOV	L,A
	STC
	CMC
	JMP	PUTARG
;
STKAND	CALL	GETARGS
	MOV	A,D
	ANA	H
	MOV	H,A
	MOV	A,E
	ANA	L
	MOV	L,A
	STC
	CMC
	JMP	PUTARG
;
; Relative operators. >,>=,<,<=,!=,=.
; Remember, ARG1 is in HL, ARG2 is in DE.  If the 
; condition is true, result is 0FFFFH (-1), otherwise,
; result is 0.
;
STKLE	CALL	GETARGS		;ARG1<=ARG2?
	XCHG
	JMP	STKLE1
STKGE	CALL	GETARGS		;ARG1>=ARG2?
STKLE1	CALL	DCMP
	JNC	STKTRUE
	JMP	STKFALSE
;
STKGT	CALL	GETARGS		;ARG1>ARG2?
	XCHG
	JMP	STKLT1
STKLT	CALL	GETARGS		;ARG1<ARG2?
STKLT1	CALL	DCMP
	JC	STKTRUE
	JMP	STKFALSE
;
STKEQ	CALL	GETARGS		;ARG1=ARG2?
	CALL	DCMP
	JZ	STKTRUE
	JMP	STKFALSE
;
STKNEQ	CALL	GETARGS 	;ARG1!=ARG2?
	CALL	DCMP
	JNZ	STKTRUE
	JMP	STKFALSE
;
STKTRUE LXI	H,-1
	XRA	A
	JMP	PUTARG
STKFALSE LXI	H,0
	XRA	A
	JMP	PUTARG
;
;STKCHOP.  Pops off all but one of the top group
;of elements of the stack.  A group is delimited
;by a count element on top of the stack.  The
;count tells how many elements below it are grouped
;together.  STKCHOP pops N-1 things off VSTACK if N was
;the old count element.
;
STKCHOP CALL	STKPOP
	JC	STKCH2		;Got to have stuff on stack!
	DCX	D		;Want to pop only N-1 things.
STKCH1	PUSH	D
	CALL	STKPOP		;Throw away one thing.
	POP	D
	JC	STKCH2		;If no stuff, big trouble!
	MOV	A,D
	ORA	E
	RZ			;Keep going until count is 0
	DCX	D
	JMP	STKCH1
STKCH2	LXI	D,0503H		;## VSTACK GROUP UNDERFLOW ##
	STC
	RET
